@@ -40,6 +40,13 @@ using NpgsqlTypes;
 
 namespace Npgsql
 {
+    /// <summary>
+    /// Represents the method that allows the application to provide a certificate collection to be used for SSL clien authentication
+    /// </summary>
+    /// <param name="certificates">A <see cref="System.Security.Cryptography.X509Certificates.X509CertificateCollection">X509CertificateCollection</see> to be filled with one or more client certificates.</param>
+    public delegate void ProvideClientCertificatesCallback(X509CertificateCollection certificates);
+
+
     /// <summary>
     /// !!! Helper class, for compilation only.
     /// Connector implements the logic for the Connection Objects to
@@ -61,6 +68,11 @@ namespace Npgsql
         /// </summary>
         internal event NotificationEventHandler Notification;
 
+        /// <summary>
+        /// Called to provide client certificates for SSL handshake.
+        /// </summary>
+        internal event ProvideClientCertificatesCallback ProvideClientCertificatesCallback;
+
         /// <summary>
         /// Mono.Security.Protocol.Tls.CertificateSelectionCallback delegate.
         /// </summary>
@@ -338,13 +350,12 @@ namespace Npgsql
             {
                 // Here we use a fake NpgsqlCommand, just to send the test query string.
                 
-                Query(new NpgsqlCommand("select 1 as ConnectionTest", this));
-                
+                //Query(new NpgsqlCommand("select 1 as ConnectionTest", this));
+                new NpgsqlCommand("select 1", this).ExecuteScalar();
                 // Clear mediator.
                 Mediator.ResetResponses();
                 this.RequireReadyForQuery = true;
                 
-                
             }
             catch
             {
@@ -470,6 +481,17 @@ namespace Npgsql
             }
         }
 
+        /// <summary>
+        /// Default SSL ProvideClientCertificatesCallback implementation.
+        /// </summary>
+        internal void DefaultProvideClientCertificatesCallback(X509CertificateCollection certificates)
+        {
+            if (ProvideClientCertificatesCallback != null)
+            {
+                ProvideClientCertificatesCallback(certificates);
+            }
+        }
+
         /// <summary>
         /// Version of backend server this connector is connected to.
         /// </summary>
@@ -946,13 +968,27 @@ namespace Npgsql
                     return paramValue;
                 }
             }
-            catch(NpgsqlException ne)
+            
+            /*
+             * In case of problems with the command above, we simply return null in order to 
+             * say we don't support it.
+             */
+              
+            catch(NpgsqlException e)
+            {
+                return null;
+            }
+            /*
+             * Original catch handler by Jon Hanna.
+             * 7.3 version doesn't support error code. Only 7.4+
+             
+             * catch(NpgsqlException ne)
             {
                 if(ne.Code == "42704")//unrecognized configuration parameter
                     return null;
                 else
                     throw;
-            }
+            }*/
         }
         private bool CheckStringConformanceRequirements()
         {
