@@ -198,7 +198,7 @@ namespace NpgsqlTypes
 		/// <param name="seconds">Number of seconds.</param>
 		/// <param name="milliseconds">Number of milliseconds.</param>
 		public NpgsqlInterval(int days, int hours, int minutes, int seconds, int milliseconds)
-			: this(0, days, new TimeSpan(hours, minutes, seconds, milliseconds).Ticks)
+			: this(0, days, new TimeSpan(0, hours, minutes, seconds, milliseconds).Ticks)
 		{
 		}
 
@@ -213,7 +213,7 @@ namespace NpgsqlTypes
 		/// <param name="seconds">Number of seconds.</param>
 		/// <param name="milliseconds">Number of milliseconds.</param>
 		public NpgsqlInterval(int months, int days, int hours, int minutes, int seconds, int milliseconds)
-			: this(months, days, new TimeSpan(hours, minutes, seconds, milliseconds).Ticks)
+			: this(months, days, new TimeSpan(0, hours, minutes, seconds, milliseconds).Ticks)
 		{
 		}
 
@@ -230,7 +230,7 @@ namespace NpgsqlTypes
 		/// <param name="seconds">Number of seconds.</param>
 		/// <param name="milliseconds">Number of milliseconds.</param>
 		public NpgsqlInterval(int years, int months, int days, int hours, int minutes, int seconds, int milliseconds)
-			: this(years*12 + months, days, new TimeSpan(hours, minutes, seconds, milliseconds).Ticks)
+			: this(years*12 + months, days, new TimeSpan(0, hours, minutes, seconds, milliseconds).Ticks)
 		{
 		}
 
@@ -293,7 +293,7 @@ namespace NpgsqlTypes
 		/// </summary>
 		public int Microseconds
 		{
-			get { return (int) (_ticks/10)%1000000; }
+			get { return (int) ((_ticks/10)%1000000); }
 		}
 
 		/// <summary>
@@ -859,7 +859,7 @@ namespace NpgsqlTypes
 						default:
 							hours = int.Parse(parts[0]);
 							minutes = int.Parse(parts[1]);
-							seconds = decimal.Parse(parts[2]);
+							seconds = decimal.Parse(parts[2], System.Globalization.CultureInfo.InvariantCulture.NumberFormat);
 							break;
 					}
                     if (isNegative)
@@ -947,16 +947,10 @@ namespace NpgsqlTypes
 				{
 					sb.Append('+');
 				}
-				sb.Append(Math.Abs(Hours).ToString("D2")).Append(':').Append(Math.Abs(Minutes).ToString("D2")).Append(':').Append(Math.Abs(Seconds).ToString("D2"));
-				long remainingTicks = Math.Abs(Ticks)%TicksPerSecond;
-				if (remainingTicks != 0)
-				{
-					while (remainingTicks%10 == 0)
-					{
-						remainingTicks /= 10;
-					}
-					sb.Append('.').Append(remainingTicks);
-				}
+                // calculate total seconds and then subtract total whole minutes in seconds to get just the seconds and fractional part
+                decimal seconds = _ticks / (decimal)TicksPerSecond - (_ticks / TicksPerMinute) * 60;
+                sb.Append(Math.Abs(Hours).ToString("D2")).Append(':').Append(Math.Abs(Minutes).ToString("D2")).Append(':').Append(Math.Abs(seconds).ToString("0#.######", System.Globalization.CultureInfo.InvariantCulture.NumberFormat));
+
 			}
 			if (sb[sb.Length - 1] == ' ')
 			{
@@ -1117,10 +1111,30 @@ namespace NpgsqlTypes
 
 		public static NpgsqlDate Parse(string str)
 		{
+
+            
 			if (str == null)
 			{
 				throw new ArgumentNullException("str");
 			}
+
+
+
+            // Handle -infinity and infinity special values.
+
+
+
+            if (str == "-infinity")
+
+                return new NpgsqlDate(DateTime.MinValue);
+
+
+
+            if (str == "infinity")
+
+                return new NpgsqlDate(DateTime.MaxValue);
+
+
 			str = str.Trim();
 			try
 			{
@@ -1614,7 +1628,7 @@ namespace NpgsqlTypes
 						seconds = int.Parse(parts[2]);
 						break;
 				}
-				int totalSeconds = hours*60*60 + minutes*60 + seconds*(neg ? -1 : 1);
+				int totalSeconds = (hours*60*60 + minutes*60 + seconds)*(neg ? -1 : 1);
 				return new NpgsqlTimeZone(totalSeconds*NpgsqlInterval.TicksPerSecond);
 			}
 			catch (OverflowException)
@@ -1825,7 +1839,7 @@ namespace NpgsqlTypes
 		/// </summary>
 		public int Microseconds
 		{
-			get { return (int) (_ticks/10)%1000000; }
+			get { return (int) ((_ticks/10)%1000000); }
 		}
 
 		/// <summary>
@@ -1890,19 +1904,12 @@ namespace NpgsqlTypes
 		}
 
 		public override string ToString()
-		{
+        {
+            // calculate total seconds and then subtract total whole minutes in seconds to get just the seconds and fractional part
+            decimal seconds = _ticks / (decimal)NpgsqlInterval.TicksPerSecond - (_ticks / NpgsqlInterval.TicksPerMinute) * 60;
 			StringBuilder sb =
 				new StringBuilder(Hours.ToString("D2")).Append(':').Append(Minutes.ToString("D2")).Append(':').Append(
-					Seconds.ToString("D2"));
-			long remainingTicks = Math.Abs(Ticks)%NpgsqlInterval.TicksPerSecond;
-			if (remainingTicks != 0)
-			{
-				while (remainingTicks%10 == 0)
-				{
-					remainingTicks /= 10;
-				}
-				sb.Append('.').Append(remainingTicks);
-			}
+                    seconds.ToString("0#.######", System.Globalization.CultureInfo.InvariantCulture.NumberFormat));
 			return sb.ToString();
 		}
 
@@ -1930,7 +1937,7 @@ namespace NpgsqlTypes
 					default:
 						hours = int.Parse(parts[0]);
 						minutes = int.Parse(parts[1]);
-						seconds = decimal.Parse(parts[2]);
+                        seconds = decimal.Parse(parts[2], System.Globalization.CultureInfo.InvariantCulture.NumberFormat);
 						break;
 				}
 				if (hours < 0 || hours > 24 || minutes < 0 || minutes > 59 || seconds < 0m || seconds >= 60 ||
@@ -2049,10 +2056,22 @@ namespace NpgsqlTypes
 			return new TimeSpan(time.Ticks);
 		}
 
-		public static explicit operator NpgsqlTime(TimeSpan interval)
-		{
-			return new NpgsqlTime(interval);
-		}
+        public static explicit operator DateTime(NpgsqlTime time)
+        {
+            try
+            {
+                return new DateTime(time.Ticks, DateTimeKind.Unspecified);
+            }
+            catch
+            {
+                throw new InvalidCastException();
+            }
+        }
+
+        public static explicit operator NpgsqlTime(TimeSpan interval)
+        {
+            return new NpgsqlTime(interval);
+        }
 
 		public NpgsqlTime AddTicks(long ticksAdded)
 		{
@@ -2479,7 +2498,13 @@ namespace NpgsqlTypes
 		public static explicit operator TimeSpan(NpgsqlTimeTZ time)
 		{
 			return (TimeSpan) time.LocalTime;
-		}
+        }
+
+        public static explicit operator DateTime(NpgsqlTimeTZ time)
+        {
+            // LocalTime property is actually time local to TimeZone
+            return new DateTime(time.AtTimeZone(NpgsqlTimeZone.CurrentTimeZone).Ticks, DateTimeKind.Local);
+        }
 	}
 
 	[Serializable]
